(* Abstract Syntax Tree and functions for printing it *)

type op = Add | Sub | Mult | Div | Mod | Equal | Neq | Less | Leq | Greater | Geq |
          And | Or

type uop = Neg | Not

type uopment = Increment | Decrement 

(* HERE ADD NUM TYPE FOR FLOATS AND INTS*)
type typ = Int | Float | Bool | MyString | Void 
	  | Vector of typ * int list

type bind = typ * string

type expr =
    Literal of int
  | Fliteral of float
  | BoolLit of bool
  | MyStringLit of string
  | Vector_lit of expr list
  | Id of string
  | Vdecl of bind 
  | Binop of expr * op * expr
  | Unop of uop * expr
  | Incrementer of expr * uopment
  | Assign of expr * expr
  | Call of string * expr list
  | Vector_access of string * expr list
  | Reference of string 
  | Dimlist of string 
  | Noexpr

type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | While of expr * stmt

type func_decl = {
    typ : typ;
    fname : string;
    formals : bind list;
    locals : bind list;
    body : stmt list;
  }

type program = bind list * func_decl list

(* Pretty-printing functions *)

let rec string_of_typ = function
    Int -> "int"
  | Float -> "float"
  | Bool -> "bool"
  | MyString -> "string" 
  | Void -> "void"
  | Vector(t, l) -> (string_of_typ t)^"["^
    (List.fold_left (fun str n -> str ^", "^string_of_int n) "" l)^"]"

let string_of_op = function
    Add -> "+"
  | Sub -> "-"
  | Mult -> "*"
  | Div -> "/"
  | Mod -> "%"
  | Equal -> "=="
  | Neq -> "!="
  | Less -> "<"
  | Leq -> "<="
  | Greater -> ">"
  | Geq -> ">="
  | And -> "&&"
  | Or -> "||"

let string_of_uop = function
    Neg -> "-"
  | Not -> "!"

let string_of_uopment = function
    Increment -> "++"
  | Decrement -> "--"

let rec string_of_expr = function
    Literal(l) -> string_of_int l
  | Fliteral(f) -> string_of_float f
  | BoolLit(true) -> "true"
  | BoolLit(false) -> "false"
  | MyStringLit(s) -> s
  | Vector_lit(el) -> "[" (* ^ String.concat ", " (List.map string_of_expr el) ^ "]" *)
  | Id(s) -> s
  | Vdecl(t, s) -> string_of_typ t^" "^s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_expr e2
  | Unop(o, e) -> string_of_uop o ^ string_of_expr e
  | Incrementer(e, uop) -> string_of_expr e ^" "^ string_of_uopment uop
  | Assign(v, e) -> string_of_expr v ^ " = " ^ string_of_expr e 
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Vector_access(v, i) ->
      v ^ "[" ^ 
      (List.fold_left (fun str n -> str ^", "^string_of_expr n) "" i) ^"]"
  | Dimlist(v) -> 
      v ^ ".dims"
  | Noexpr -> ""
  (*ADD PATTERN FOR VECTOR_LIT AND VECTOR_ACCESS*)

let rec string_of_stmt = function
    Block(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Expr(expr) -> string_of_expr expr ^ ";\n";
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n" ^ string_of_stmt s
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n" ^
      string_of_stmt s1 ^ "else\n" ^ string_of_stmt s2
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3  ^ ") " ^ string_of_stmt s
  | While(e, s) -> "while (" ^ string_of_expr e ^ ") " ^ string_of_stmt s

let string_of_vdecl (t, id) = string_of_typ t ^ " " ^ id ^ ";\n"

let string_of_fdecl fdecl =
  string_of_typ fdecl.typ ^ " " ^
  fdecl.fname ^ "(" ^ String.concat ", " (List.map snd fdecl.formals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_vdecl fdecl.locals) ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdecl vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
